//
// NDR.cpp
//

#include "stdafx.h"
#include "NDR.h"
#include "ProxyFileInfoRetriever.h"

//
// Definitions.
//

static const GUID g_guidNull =
	{ 0x0000000, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } };

#define FC_AUTO_HANDLE			0x33

#define Oi_FULL_PTR_USED					0x01 // Uses the full pointer package. 
#define Oi_RPCSS_ALLOC_USED					0x02 // Uses the RpcSs memory package. 
#define Oi_OBJECT_PROC						0x04 // A procedure in an object interface. 
#define Oi_HAS_RPCFLAGS						0x08 // The procedure has nonzero Rpc flags. 
#define Oi_USE_NEW_INIT_ROUTINES			0x40 // Uses Windows NT3.5 Beta2+ init routines 
#define ENCODE_IS_USED						0x10 // Used only in pickling. 
#define DECODE_IS_USED						0x20 // Used only in pickling. 
#define Oi_IGNORE_OBJECT_EXCEPTION_HANDLING	0x10 // Not used anymore (old OLE). 
#define Oi_HAS_COMM_OR_FAULT				0x20 // In raw RPC only, [comm _, fault_status]. 
#define Oi_OBJ_USE_V2_INTERPRETER			0x20 // In DCOM only, use Oif interpreter. 

typedef struct
{
  unsigned char   ServerMustSize      : 1;    // 0x01
  unsigned char   ClientMustSize      : 1;    // 0x02
  unsigned char   HasReturn           : 1;    // 0x04
  unsigned char   HasPipes            : 1;    // 0x08
  unsigned char   Unused              : 1;
  unsigned char   HasAsyncUuid        : 1;    // 0x20
  unsigned char   HasExtensions       : 1;    // 0x40
  unsigned char   HasAsyncHandle      : 1;    // 0x80
} INTERPRETER_OPT_FLAGS, *PINTERPRETER_OPT_FLAGS;

typedef struct
{
  unsigned short  MustSize            : 1;    // 0x0001
  unsigned short  MustFree            : 1;    // 0x0002
  unsigned short  IsPipe              : 1;    // 0x0004
  unsigned short  IsIn                : 1;    // 0x0008
  unsigned short  IsOut               : 1;    // 0x0010
  unsigned short  IsReturn            : 1;    // 0x0020
  unsigned short  IsBasetype          : 1;    // 0x0040
  unsigned short  IsByValue           : 1;    // 0x0080
  unsigned short  IsSimpleRef         : 1;    // 0x0100
  unsigned short  IsDontCallFreeInst  : 1;    // 0x0200
  unsigned short  SaveForAsyncFinish  : 1;    // 0x0400
  unsigned short  Unused              : 2;
  unsigned short  ServerAllocSize     : 3;    // 0xe000
} PARAM_ATTRIBUTES, *PPARAM_ATTRIBUTES;

#define FC_BYTE				0x01
#define FC_CHAR				0x02
#define FC_SMALL			0x03
#define FC_USMALL			0x04
#define FC_WCHAR			0x05
#define FC_SHORT			0x06
#define FC_USHORT			0x07
#define FC_LONG				0x08
#define FC_ULONG			0x09
#define FC_FLOAT			0x0a
#define FC_HYPER			0x0b
#define FC_DOUBLE			0x0c
#define FC_ENUM16			0x0d
#define FC_ENUM32			0x0e
#define FC_ERROR_STATUS_T	0x10
#define FC_INT3264			0xb8
#define FC_UINT3264			0xb9

#define FC_RP					0x11 
#define FC_UP					0x12
#define FC_FP					0x14
#define FC_OP					0x13
#define FC_IP					0x2f
#define FC_BYTE_COUNT_POINTER	0x2c

#define FC_CONSTANT_IID		0x5a

#define FC_PAD				0x5c

#define FC_ALLOCATE_ALL_NODES	0x01
#define FC_DONT_FREE			0x02
#define FC_ALLOCED_ON_STACK		0x04
#define FC_SIMPLE_POINTER		0x08
#define FC_POINTER_DEREF		0x10

// other definitions (from Internet/from disassembling MidlC.exe...)...
#define FC_SMFARRAY					0x1d
#define FC_CARRAY					0x1b
#define FC_CVARRAY					0x1c
#define FC_BOGUS_ARRAY				0x21
#define FC_C_CSTRING				0x22
#define FC_C_WSTRING				0x25
#define FC_STRING_SIZED				0x44
#define FC_STRUCT					0x15
#define FC_PSTRUCT					0x16
#define FC_CSTRUCT					0x17
#define FC_BOGUS_STRUCT				0x1a
#define FC_ENCAPSULATED_UNION		0x2a
#define FC_NON_ENCAPSULATED_UNION	0x2b
#define FC_USER_MARSHAL				0xb4
#define FC_LGFARRAY					0x1e
#define FC_SMVARRAY					0x1f
#define FC_LGVARRAY					0x20
#define FC_CSTRING					0x26
#define FC_WSTRING					0x29
#define FC_SSTRING					0x28
#define FC_C_SSTRING				0x24
#define FC_CPSTRUCT					0x18
#define FC_CVSTRUCT					0x19
#define FC_HARD_STRUCTURE			0xb1
#define FC_TRANSMIT_AS				0x2d
#define FC_REPRESENT_AS				0x2e
#define FC_RANGE					0xb7

//
// GetReport Implementation.
//

static BOOL GetInterfaceName( IN CHAR* szGUID, OUT BOOL& bTypeLib, OUT charstring& csName, OUT charstring& csErr )
{
	bTypeLib = FALSE;
	csName = "";
	csErr = "";

	BOOL		bRetVal = FALSE;

	// read from the registry.

	LONG		lRet, lRet2;

	HKEY		hInt = NULL;
	lRet = ::RegOpenKey( HKEY_CLASSES_ROOT, "Interface", & hInt );
	if ( lRet == ERROR_SUCCESS )
	{
		charstring	csGUID = szGUID;

		GUID		guid;
		BOOL		bGuidRes = ::CharStrToGuid( szGUID, & guid );
		if ( bGuidRes && csGUID.size() >= 3 )
		{
			if ( csGUID[ 0 ] == '{' && csGUID[ csGUID.size() - 1 ] == '}' )
				csGUID = csGUID.substr( 1, csGUID.size() - 2 );

			HKEY		hSub1 = NULL, hSub2 = NULL;
			lRet = ::RegOpenKey( hInt, csGUID.c_str(), & hSub1 );
			lRet2 = ::RegOpenKey( hInt, ( "{" + csGUID + "}" ).c_str(), & hSub2 );

			if ( lRet == ERROR_SUCCESS || lRet2 == ERROR_SUCCESS )
			{
				// query value.

				DWORD		dwType = 0;

				CHAR		szBuffer[ 1 * 1024 ];
				DWORD		dwDataLen = sizeof( szBuffer );

				lRet = ::RegQueryValueEx(
					hSub1 ? hSub1 : hSub2,
					NULL, NULL, & dwType, (LPBYTE) szBuffer, & dwDataLen );

				if ( lRet == ERROR_SUCCESS && dwDataLen && dwType == REG_SZ )
				{
					csName = szBuffer;
					bRetVal = TRUE;
				}
				else
				{
					csErr = "ERROR (Unable to read the name of interface from registry !)";
				}

				// check to see whether there is a TYPELIB.

				HKEY		hTypeLib = NULL;

				lRet = ::RegOpenKey(
					hSub1 ? hSub1 : hSub2,
					"TypeLib", & hTypeLib );

				if ( lRet == ERROR_SUCCESS )
				{
					bTypeLib = TRUE;
					::RegCloseKey( hTypeLib );
				}

				// close key.

				if ( hSub1 )
					::RegCloseKey( hSub1 );
				if ( hSub2 )
					::RegCloseKey( hSub2 );
			}
			else
			{
				csErr = "ERROR (Error when opening interface subkey ! [ name : \"" + csGUID + "\" ])";
			}
		}
		else
		{
			csErr = "ERROR (Invalid IID)";
		}

		::RegCloseKey( hInt );
	}
	else
	{
		csErr = "ERROR (Unable to open \"Interface\" key in CLASSES_ROOT !)";
	}

	// return.

	return bRetVal;
}

static VOID PrintBaseType( IN BYTE bTypeFC, OUT charstring& csP )
{
	// print.

	switch( bTypeFC )
	{
		case FC_BYTE:
			csP += "BYTE ";
			break;
		case FC_CHAR:
			csP += "CHAR ";
			break;
		case FC_SMALL:
			csP += "SMALL ";
			break;
		case FC_USMALL:
			csP += "USMALL ";
			break;
		case FC_WCHAR:
			csP += "WCHAR ";
			break;
		case FC_SHORT:
			csP += "SHORT ";
			break;
		case FC_USHORT:
			csP += "USHORT ";
			break;
		case FC_LONG:
			csP += "LONG ";
			break;
		case FC_ULONG:
			csP += "ULONG ";
			break;
		case FC_FLOAT:
			csP += "FLOAT ";
			break;
		case FC_HYPER:
			csP += "HYPER ";
			break;
		case FC_DOUBLE:
			csP += "DOUBLE ";
			break;
		case FC_ENUM16:
			csP += "ENUM16 ";
			break;
		case FC_ENUM32:
			csP += "ENUM32 ";
			break;
		case FC_ERROR_STATUS_T:
			csP += "ERROR_STATUS ";
			break;
		case FC_INT3264:
			csP += "INT3264 ";
			break;
		case FC_UINT3264:
			csP += "UINT3264 ";
			break;
		default:
			csP += "??? ";
			break;
	}

	// return.

	return;
}

static VOID PrintComplexType( IN BYTE* pbTypeFormatTbl, IN USHORT usTypTblRefPos, IN USHORT usTypeOff, OUT charstring& csP )
{
	// convert the off to absolute.

	usTypeOff = usTypTblRefPos + usTypeOff;

	// begin reading.

	BYTE*		pbT = & pbTypeFormatTbl[ usTypeOff ];

	BYTE		bType = * pbT ++;
	switch( bType )
	{
	case FC_RP: // A reference pointer.
	case FC_UP: // A unique pointer.
	case FC_FP: // A full pointer.
	case FC_OP: // A unique pointer in an object interface.
		{
			//
			// pointer.
			//

			BYTE		bAttr = * pbT ++;

			if ( bAttr & FC_SIMPLE_POINTER )
			{
				BYTE		bSimType = * pbT ++;
				pbT ++; // skip FC_PAD...

				// print.

				charstring		csRES = "";
				::PrintBaseType( bSimType, csRES );
				::TrimAtBothSides( csRES );
				csP += "SIMPLE_PTR{" + csRES + "} ";
			}
			else
			{
				USHORT		usTypTblRefPos = pbT - pbTypeFormatTbl;
				USHORT		usComplexOff = * ( USHORT* ) pbT;
				pbT += sizeof( USHORT );

				// print.

				charstring		csRES = "";
				::PrintComplexType( pbTypeFormatTbl, usTypTblRefPos, usComplexOff, csRES );
				::TrimAtBothSides( csRES );
				csP += "PTR_TO_TYPE{" + csRES + "} ";
			}
		}
		break;

	case FC_IP: // A interface pointer.
		{
			//
			// interface pointer.
			//

			BYTE			bPtrType = * pbT ++;

			switch( bPtrType )
			{
			case FC_CONSTANT_IID: // A pointer to a known IID.
				{
					GUID*			pGuid = (GUID*) pbT;
					pbT += sizeof( GUID );

					// print.

					char		szGUID[ MAX_PATH ];

					::sprintf( szGUID,"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
						pGuid->Data1,
						pGuid->Data2,
						pGuid->Data3,
						pGuid->Data4[0],
						pGuid->Data4[1],
						pGuid->Data4[2],
						pGuid->Data4[3],
						pGuid->Data4[4],
						pGuid->Data4[5],
						pGuid->Data4[6],
						pGuid->Data4[7] );

					BOOL		bTypeLib = FALSE;
					charstring	csIntName = "";
					charstring	csIntErr = "";
					BOOL		bIntRes = ::GetInterfaceName( szGUID, bTypeLib, csIntName, csIntErr );

					csP += "IID{" +
						( bIntRes ? csIntName : charstring( szGUID ) )
						+ "} ";
				}
				break;

			case FC_PAD: // A pointer to a iid_id() IID.
				{
					/* DEBUG: not implemented! */ csP += "PTR_TO_IID_ID() ";
				}
				break;

			default:
				{
					/* DEBUG: not implemented! */ csP += "FC_IP_??? ";
				}
				break;
			}
		}
		break;

	case FC_BYTE_COUNT_POINTER:
		{
			//
			// [byte_count] pointer.
			//

			BYTE		bBCType = * pbT ++;
			switch( bBCType )
			{
			case FC_PAD: // format #2.
				{
					// follows: byte_count_description<> and pointee_description<> ...

					/* DEBUG: not implemented! */ csP += "FC_BYTE_COUNT_PTR_#2 ";
				}
				break;

			default: // format #1.
				{
					BYTE		bSimType = * pbT ++;
					// follows: byte_count_description<> ...

					/* DEBUG: not implemented! */ csP += "FC_BYTE_COUNT_PTR_#1 ";
				}
				break;
			}
		}
		break;

	case FC_SMFARRAY: // fixed-(16bit in)size array. (block-copyable)
	case FC_LGFARRAY: // fixed-(32bit in)size array. (block-copyable)
		{
			BYTE		bAlign = * pbT ++;

			ULONG		lTotSize = ( bType == FC_SMFARRAY ) ? * (USHORT*) pbT : * (ULONG*) pbT;
			pbT += ( bType == FC_SMFARRAY ) ? sizeof( USHORT ) : sizeof( ULONG );

			// follows: [pointer_layout<>], element_description<> and FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_FARRAY ";
		}
		break;

	case FC_CARRAY: // A Conformant Array.
	case FC_CVARRAY: // A Conformant Varying Array;
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usElSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: conformance_description<>,
			// ( if FC_CVARRAY then variance_description<> ),
			// [pointer_layout<>], element_description<> and FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_C[V]ARRAY ";
		}
		break;

	case FC_SMVARRAY: // A small varying array.
	case FC_LGVARRAY: // A large varying array.
		{
			BYTE		bAlign = * pbT ++;

			ULONG		lTotSize = ( bType == FC_SMVARRAY ) ? * (USHORT*) pbT : * (ULONG*) pbT;
			pbT += ( bType == FC_SMVARRAY ) ? sizeof( USHORT ) : sizeof( ULONG );

			ULONG		lNumEl = ( bType == FC_SMVARRAY ) ? * (USHORT*) pbT : * (ULONG*) pbT;
			pbT += ( bType == FC_SMVARRAY ) ? sizeof( USHORT ) : sizeof( ULONG );

			USHORT		usElSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: variance_description<>, [pointer_layout<>], element_description<> and FC_END.

			/* DEBUG: not implemented! */ csP += "FC_VARRAY ";
		}
		break;

	case FC_BOGUS_ARRAY: // A complex array.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usNumEl = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: conformance_description<>, variance_description<>, element_description<> and FC_END.

			/* DEBUG: not implemented! */ csP += "FC_BOGUS_ARRAY ";
		}
		break;

	case FC_CSTRING: // A Nonconformant String.
	case FC_WSTRING: // A Nonconformant String.
		{
			pbT += sizeof( BYTE ); // skip FC_PAD.

			USHORT		usStrSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			/* DEBUG: not implemented! */ csP += "FC_[C/W]STRING ";
		}
		break;

	case FC_C_CSTRING: // A Conformant String.
	case FC_C_WSTRING: // A Conformant String.
		{
			BYTE		bStrType = * pbT ++;
			switch( bStrType )
			{
			case FC_PAD:
				{
					/* DEBUG: not implemented! */ csP += "FC_C_[C/W]STRING_#1 ";
				}
				break;

			case FC_STRING_SIZED:
				{
					// follows: conformance_description<> ...

					/* DEBUG: not implemented! */ csP += "FC_C_[C/W]STRING_#2 ";
				}
				break;

			default:
				{
					/* DEBUG: not implemented! */ csP += "FC_C_[C/W]STRING_??? ";
				}
				break;
			}
		}
		break;

	case FC_SSTRING: // A Nonconformant Structure String.
		{
			BYTE		bElSize = * pbT ++;

			USHORT		usNumEl = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			/* DEBUG: not implemented! */ csP += "FC_SSTRING ";
		}
		break;

	case FC_C_SSTRING: // A Conformant Structure String.
		{
			BYTE		bElSize = * pbT ++;

			BYTE		bStrType = * pbT ++;
			switch( bStrType )
			{
			case FC_STRING_SIZED:
				{
					pbT += sizeof( BYTE ); // skip FC_PAD ...

					// follows: conformance_description<> ...

					/* DEBUG: not implemented! */ csP += "FC_C_SSTRING_#1 ";
				}
				break;

			default:
				{
					pbT --; // restore type byte.

					/* DEBUG: not implemented! */ csP += "FC_C_SSTRING_#2 ";
				}
				break;
			}
		}
		break;

	case FC_STRUCT: // A Simple Structure.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: member_layout<> and FC_END.

			/* DEBUG: not implemented! */ csP += "FC_STRUCT ";
		}
		break;

	case FC_PSTRUCT: // A Simple Structure with Pointers.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: pointer_layout<>, member_layout<> and FC_END.

			/* DEBUG: not implemented! */ csP += "FC_PSTRUCT ";
		}
		break;

	case FC_CSTRUCT: // A Conformant Structure.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2ArrDesc = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: member_layout<>, FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_CSTRUCT ";
		}
		break;

	case FC_CPSTRUCT: // A Conformant Structure with Pointers.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2ArrDesc = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: pointer_layout<>, member_layout<> and FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_CPSTRUCT ";
		}
		break;

	case FC_CVSTRUCT: // A Conformant Varying Structure (with or without Pointers).
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2ArrDesc = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: [pointer_layout<>], layout<> and FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_CVSTRUCT ";
		}
		break;

	case FC_HARD_STRUCTURE: // A Hard Structure.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			pbT += sizeof( DWORD ); // skip reserved ...

			USHORT		usEnumOff = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usCopySz = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usMemCpyInc = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usUnionDescOff = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: member_layout<> and FC_END ...

			/* DEBUG: not implemented! */ csP += "FC_HARD_STRUCT ";
		}
		break;

	case FC_BOGUS_STRUCT: // A Complex Structure.
		{
			BYTE		bAlign = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2ConfArrDesc = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2PtrLay = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: member_layout<>, FC_END and [pointer_layout<>] ...

			/* DEBUG: not implemented! */ csP += "FC_BOGUS_STRUCT ";
		}
		break;

	case FC_ENCAPSULATED_UNION: // A Encapsulated Union.
		{
			BYTE		bSwitchTyp = * pbT ++;

			USHORT		usMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			// follows: union_arm_selector<> ...

			/* DEBUG: not implemented! */ csP += "FC_ENCAPS_UNION ";
		}
		break;

	case FC_NON_ENCAPSULATED_UNION: // A Nonencapsulated Union.
		{
			BYTE		bSwitchTyp = * pbT ++;

			// follows: switch_is_description<> and offset_to_size_and_arm_description<2> ...

			/* DEBUG: not implemented! */ csP += "FC_NON_ENCAPS_UNION ";
		}
		break;

	case FC_TRANSMIT_AS: // A transmit-as desc.
	case FC_REPRESENT_AS: // A represent-as desc.
		{
			BYTE		bFlags = * pbT ++;

			USHORT		usQuiInd = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usPreTypMemSiz = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usTraTypBufSiz = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usTraTypOff = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			/* DEBUG: not implemented! */ csP += "FC_TRANSMIT/REPRESENT_AS ";
		}
		break;

	case FC_USER_MARSHAL: // A User Marshal desc.
		{
			BYTE		bFlags = * pbT ++;

			USHORT		usQuaInd = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usUserTypMemSize = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usTranTypBufSiz = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			USHORT		usOff2TransTyp = * (USHORT*) pbT;
			pbT += sizeof( USHORT );

			/* DEBUG: not implemented! */ csP += "FC_USER_MARSHAL ";
		}
		break;

	case FC_RANGE: // A [range] desc.
		{
			BYTE		bFlags = * pbT ++;

			ULONG		ulLowVal = * (ULONG*) pbT;
			pbT += sizeof( ULONG );

			ULONG		ulHighVal = * (ULONG*) pbT;
			pbT += sizeof( ULONG );

			/* DEBUG: not implemented! */ csP += "FC_RANGE ";
		}
		break;

	default:
		{
			/* DEBUG: not implemented! */ csP += "FC_??? ";
		}
		break;
	}

	// return.

	return;
}

static CProxyFileInfoRetriever*		g_pRetriever = NULL;

VOID GetReport( charstring& csOut, GUID& guidInterface )
{
	// get the instance to the retriever.

	if ( g_pRetriever == NULL )
	{
		::CoInitialize( 0 );
		g_pRetriever = ::CreatePfiRetrieverInstance ();
	}

	// prologue.

	csOut += "================================================================\r\n";
	csOut += "NDR report of interface: ";

	char		szGUID[ MAX_PATH ];

	sprintf( szGUID,"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
		guidInterface.Data1,
		guidInterface.Data2,
		guidInterface.Data3,
		guidInterface.Data4[0],
		guidInterface.Data4[1],
		guidInterface.Data4[2],
		guidInterface.Data4[3],
		guidInterface.Data4[4],
		guidInterface.Data4[5],
		guidInterface.Data4[6],
		guidInterface.Data4[7] );

	csOut += charstring( "{" ) + szGUID + "}.\r\n";

	BOOL		bTypeLib = FALSE;

	csOut += "Interface human name: ";

	charstring	csName = "";
	charstring	csE = "";
	if ( ::GetInterfaceName( szGUID, bTypeLib, csName, csE ) )
		csOut += "\"" + csName + "\"";
	else
		csOut += csE;

	csOut += ".\r\n================================================================\r\n\r\n";

	// check whether a PS is specified.

	BOOL		bPS = TRUE;

	CLSID		psclsid;
	::memset( & psclsid, 0, sizeof( psclsid ) );

	HRESULT		hr;

	hr = ::CoGetPSClsid( guidInterface, & psclsid );
	if ( FAILED( hr ) || ::IsEqualGUID( psclsid, g_guidNull ) )
		bPS = FALSE;

	// get server info.

	csOut += "general info:\r\n";
	csOut += "-------------\r\n";

	MIDL_SERVER_INFO*		pServerInfo = NULL;
	ULONG					ulDispatchTableCount = 0;
	IID						iidBaseInt;
	::memset( & iidBaseInt, 0, sizeof( iidBaseInt ) );
	BYTE*					pbTypeFormatTbl = NULL;

	if ( bPS == FALSE )
	{
		csOut += "  PSDLL: proxy stub data not present.\r\n";
	}
	else
	{
		charstring				csPsDllName = "";
		charstring				csError = "";

		BOOL bGsiRes = ::GetServerInfo( g_pRetriever, guidInterface, & pServerInfo, & ulDispatchTableCount, & iidBaseInt, & pbTypeFormatTbl, csPsDllName, csError );

		if ( csPsDllName.size() )
			csOut += "  PSDLL: " + csPsDllName + "\r\n";

		csOut += "  MIDL_SERVER_INFO: ";
		if ( bGsiRes == FALSE )
		{
			csOut += "ERROR! (" + csError + ")";
		}
		else
		{
			CHAR		szNumber[ 32 ];
			::sprintf( szNumber, "%i", ulDispatchTableCount );
			csOut += charstring( "found, DispatchTableCount = " ) + szNumber + ".";
		}
		csOut += "\r\n";
	}

	csOut += "  TYPELIB: ";
	if ( bTypeLib )
		csOut += "seems to be present. Use OLEVIEW.\r\n";
	else
		csOut += "not present.\r\n";

	//
	// dump methods data.
	//

	if ( pServerInfo )
	{
		csOut += "\r\n";
		int			iStart = 3;

		// header.

		csOut += "interface methods:\r\n";
		csOut += "------------------\r\n";

		// IUnknown.

		if ( ::IsEqualGUID( iidBaseInt, g_guidNull ) == FALSE &&
			::IsEqualGUID( iidBaseInt, IID_IUnknown ) == FALSE &&
			::IsEqualGUID( iidBaseInt, IID_IDispatch ) == FALSE )
		{
			csOut += "  -WARNING- base interface iid not recognized.\r\n";
		}

		csOut += "  [proc#0] = IUnknown::QueryInterface\r\n";
		csOut += "  [proc#1] = IUnknown::AddRef\r\n";
		csOut += "  [proc#2] = IUnknown::Release\r\n";

		// IDispatch.

		if ( ::IsEqualGUID( iidBaseInt, IID_IDispatch ) )
		{
			// we will start dumping from the 7th method.

			iStart = 7;

			// standard IDispatch text will be pumped.

			csOut += "  [proc#3] = IDispatch::GetTypeInfoCount\r\n";
			csOut += "  [proc#4] = IDispatch::GetTypeInfo\r\n";
			csOut += "  [proc#5] = IDispatch::GetIDsOfNames\r\n";
			csOut += "  [proc#6] = IDispatch::Invoke\r\n";
		}

		// iterate.

		for( int i=iStart; i<ulDispatchTableCount; i ++ )
		{
			// dump the method format data.

			try
			{
				charstring		csErr = "";

				// get pointer.

				BYTE*			pbPtr = ::GetMethodNdrData( pServerInfo, i, csErr );
				if ( pbPtr == NULL )
				{
					csOut += "  ERROR (" + csErr + ")\r\n";
					continue;
				}

				// check handle_type and other proc header desc fields.

				if ( * pbPtr ++ != FC_AUTO_HANDLE )
				{
					csOut += "  ERROR ( handle_type != FC_AUTO_HANDLE )\r\n";
					continue;
				}

				BYTE		bOF = * pbPtr ++;

				if ( ( bOF & Oi_OBJ_USE_V2_INTERPRETER ) == FALSE )
				{
					csOut += "  ERROR ( only -Oif data supported )\r\n";
					continue;
				}

				if ( bOF & Oi_HAS_RPCFLAGS )
					pbPtr += sizeof( DWORD );

				USHORT		usProcNum = * (USHORT*) pbPtr;
				pbPtr += sizeof( USHORT );

				USHORT		usStackSize = * (USHORT*) pbPtr;
				pbPtr += sizeof( USHORT );

				pbPtr += sizeof( USHORT ) + sizeof( USHORT ); // skip "constant_client_buffer_size<2>" and "constant_server_buffer_size<2>".

				INTERPRETER_OPT_FLAGS*		pOi2F = (INTERPRETER_OPT_FLAGS*) pbPtr ++;

				BYTE				bNumParams = * pbPtr ++;

				BYTE				bNumEffectiveParams = bNumParams;
				if ( pOi2F->HasReturn )
					bNumEffectiveParams --;

				if ( pOi2F->HasExtensions )
				{
					BYTE			bExtVer = * pbPtr ++;
					pbPtr += bExtVer - 1;
				}

				// compose the procedure pseudo prototype.

				CHAR				szNumber[ 32 ];

				::sprintf( szNumber, "%i", usProcNum );
				charstring			csPROC = charstring( " [proc#" ) + szNumber + "] ( ";

				if ( bNumEffectiveParams == 0 )
					csPROC += "VOID";
				if ( pOi2F->HasReturn == 0 )
					csPROC = "VOID" + csPROC;

				for( int j=0; j<bNumParams; j ++ )
				{
					charstring			csP = "";

					// check param descriptor fields...

					PARAM_ATTRIBUTES*		pParamAttr = (PARAM_ATTRIBUTES*) pbPtr;
					pbPtr += sizeof( USHORT );

					USHORT					usStackOff = * (USHORT*) pbPtr;
					pbPtr += sizeof( USHORT );

					if ( pParamAttr->IsIn )
						csP += "IN ";
					if ( pParamAttr->IsOut )
						csP += "OUT ";
					if ( pParamAttr->MustSize )
						csP += "MS ";

					if ( pParamAttr->IsBasetype )
					{
						BYTE		bTypeFC = * pbPtr ++;
						pbPtr ++; // skip unused byte.

						// print base type.

						::PrintBaseType( bTypeFC, csP );
					}
					else
					{
						USHORT		usTypeOff = * (USHORT*) pbPtr;
						pbPtr += sizeof( USHORT );

						if ( pbTypeFormatTbl == NULL )
							csP += "(ERR:no types!) ";
						else
							::PrintComplexType( pbTypeFormatTbl, 0, usTypeOff, csP );
					}

					// insert name.

					::sprintf( szNumber, "%i", usStackOff );
					csP += "[#" + charstring( szNumber ) + "]"; // note: no last space...

					// insert comma.

					if ( j < bNumEffectiveParams - 1 )
						csP += ", ";

					// append.

					if ( pParamAttr->IsReturn )
						csPROC = csP + csPROC;
					else
						csPROC += csP;
				}

				// append.

				csOut += "  " + csPROC + " )\r\n";
			}
			catch( ... )
			{
				csOut += "  ERROR (access violation exception)\r\n";
				continue;
			}
		}
	}

	// return.

	return;
}
